/**
  @file main.c
 */
/*
 * ============================================================================
 * History
 * =======
 * 2017-03-15 : Created
 *
 * (C) Copyright Bridgetek Pte Ltd
 * ============================================================================
 *
 * This source code ("the Software") is provided by Bridgetek Pte Ltd
 * ("Bridgetek") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of Bridgetek liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on Bridgetek's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, Bridgetek
 * has no liability in relation to those amendments.
 * ============================================================================
 */

/* INCLUDES ************************************************************************/

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <ft900.h>
#include <ft900_uart_simple.h>

#include <ft900_usb.h>
#include <ft900_usbd.h>
#include <ft900_usb_audio.h>
#include <ft900_startup_dfu.h>

/* UART support for printf output. */
#include "tinyprintf.h"

/**
 @brief I2S device type connected.
 @details Each I2S device type supported will have different setup procedures
  and configuration settings. This macro is used to select one of the
  supported DAc types for use in this application.
  I2S_TYPE can be passed as a directive to the makefile.
  Supported DAC types are Wolfson/Cirrus WM8731.
 */
//@{
#define WM8731 1

#ifndef I2S_TYPE
#define I2S_TYPE WM8731
#endif // I2S_TYPE
//@}

/**
 @brief I2C connection.
 @details Each I2C device type supported will have different address.
  I2C_CHANNEL is either 1 or 0 depending on the I2C Master channel on the
  FT9xx that the DAC is connected to.
  */
//@{
#if I2S_TYPE == WM8731
#define WM8731_WR_ADDR (0x34)
#define WM8731_RD_ADDR (0x35)
#define I2C_CHANNEL 1
#else
#error unsupported I2S type
#endif
//@}

/**
 @brief Buffer audio stream from host.
 @details This adds a circular buffer to the data path between the
 host and the DAC. This can be used to process the audio stream if
 required or produce a delay.
 */
#define AUD_USE_BUFFER

/**
 @brief I2S Master Mode selection
 @details Define this to set the FT90x to be the I2S Master. The
 MM900EV2 boards can only be configured as I2S Slave devices. In
 this case the DAC acts as the master.
 */
#undef AUD_MASTER_MODE_FT90x

/**
 @brief Include a DFU Interface in the configuration.
 @details This adds an interface to the USB configuration descriptor
  to present a DFU option to the host. It will add in WCID descriptors
  so that a WinUSB-based device driver can be used with a DFU program
  such as dfu-util.
 */
#undef USB_INTERFACE_USE_DFU

/**
 @brief Start DFU Interface briefly before the UVC dac.
 @details This enables the DFU interface to appear for a short period
  of time to present a DFU option to the host. After this, the
  configuration will be removed and the main program's UVC configuration
  will start. A DFU program can start when the DFU configuration is
  active and update firmware.
 */
#undef USB_INTERFACE_USE_STARTUPDFU

#define BRIDGE_DEBUG
#ifdef BRIDGE_DEBUG
#define BRIDGE_DEBUG_PRINTF(...) do {tfp_printf(__VA_ARGS__);} while (0)
#else
#define BRIDGE_DEBUG_PRINTF(...)
#endif

/* For MikroC const qualifier will place variables in Flash
 * not just make them constant.
 */
#if defined(__GNUC__)
#define DESCRIPTOR_QUALIFIER const
#elif defined(__MIKROC_PRO_FOR_FT90x__)
#define DESCRIPTOR_QUALIFIER data
#endif

/* CONSTANTS ***********************************************************************/

/**
 @name USB and Hub Configuration
 @brief Indication of how the USB device is powered and the size of
 	 the control endpoint Max Packets.
 */
//@{
// USB Bus Powered - set to 1 for self-powered or 0 for bus-powered
#ifndef USB_SELF_POWERED
#define USB_SELF_POWERED 0
#endif // USB_SELF_POWERED
#if USB_SELF_POWERED == 1
#define USB_CONFIG_BMATTRIBUTES_VALUE (USB_CONFIG_BMATTRIBUTES_SELF_POWERED | USB_CONFIG_BMATTRIBUTES_RESERVED_SET_TO_1)
#else // USB_SELF_POWERED
#define USB_CONFIG_BMATTRIBUTES_VALUE USB_CONFIG_BMATTRIBUTES_RESERVED_SET_TO_1
#endif // USB_SELF_POWERED
// USB Endpoint Zero packet size (both must match)
#define USB_CONTROL_EP_MAX_PACKET_SIZE 64
#define USB_CONTROL_EP_SIZE USBD_EP_SIZE_64
//@}


/**
 @name DFU Configuration
 @brief Determines the parts of the DFU specification which are supported
        by the DFU library code. Features can be disabled if required.
 */
//@{
#define DFU_ATTRIBUTES USBD_DFU_ATTRIBUTES
//@}

/**
 @name Device Configuration Areas
 @brief Size and location reserved for string descriptors.
 Leaving the allocation size blank will make an array exactly the size
 of the string allocation.
 Note: Specifying the location is not supported by the GCC compiler.
 */
//@{
// String descriptors - allow a maximum of 256 bytes for this
#define STRING_DESCRIPTOR_LOCATION 0x80
#define STRING_DESCRIPTOR_ALLOCATION 0x100
//@}

/**
 @name DFU_TRANSFER_SIZE definition.
 @brief Number of bytes in block, sent in each DFU_DNLOAD request
 from the DFU update program on the host. This is simplified
 in that the meaning of a block is an arbitrary number of
 bytes. This is intentionally a multiple of the maximum
 packet size for the control endpoints.
 It is used in the DFU functional descriptor as wTransferSize.
 The maximum size supported by the DFU library is 256 bytes
 which is the size of a page of Flash.
 */
//@{
#define DFU_TRANSFER_SIZE USBD_DFU_MAX_BLOCK_SIZE
#define DFU_TIMEOUT USBD_DFU_TIMEOUT
//@}

/**
 @name USB_PID_AUDIO configuration.
 @brief Run Time Product ID for UVC function.
 */
//@{
#define USB_PID_AUDIO 0x0fd5
//@}

/**
 @name DFU_USB_PID_DFUMODE configuration.
 @brief FTDI predefined DFU Mode Product ID.
 */
//@{
#define DFU_USB_PID_DFUMODE 0x0fde
//@}

/**
 @name DFU_USB_INTERFACE configuration..
 @brief Run Time and DFU Mode Interface Numbers.
 */
//@{
#define DFU_USB_INTERFACE_RUNTIME 2
#define DFU_USB_INTERFACE_DFUMODE 0
//@}

/**
 @name WCID_VENDOR_REQUEST_CODE for WCID.
 @brief Unique vendor request code for WCID OS Vendor Extension validation.
 */
//@{
#define WCID_VENDOR_REQUEST_CODE	 0xF1
//@}

/**
 @brief Endpoint definitions for UAD device.
 */
//@{

/// Isochronous Audio Stream OUT endpoint
#define UAD_EP_AUDIO_STREAM				USBD_EP_1
#define UAD_EP_AUDIO_STREAM_SIZE		0x100
#define UAD_EP_AUDIO_STREAM_USBD_SIZE	USBD_EP_SIZE_256
#define UAD_EP_AUDIO_STREAM_INTERVAL_HS	4 /// 1ms polling interval
#define UAD_EP_AUDIO_STREAM_INTERVAL_FS	1 /// 1ms polling interval
//@}

/**
 * @brief Audio format definitions.
 */
//@{
/**
 * @brief Number of audio data channels.
 * @details This may be 2 for stereo or 1 for mono. Other channel
 * configurations may be set. The bitmask in AUD_CHANNEL_CONFIG
 * must match the number of channels.
 */
#define AUD_NUM_CHANNELS 2

/**
 * @brief Audio channel spacial positioning.
 */
#define AUD_CHANNEL_CONFIG (USB_AUDIO_SPACIAL_FL | /* D0: Front Left - FL */ \
							USB_AUDIO_SPACIAL_FR) /* D1: Front Right - FR */

/** @brief Resolution for samples used for encoded audio data.
 * @details This can be 16, 24 or 32 bits.
 */
#define AUD_BIT_RESOLUTION 16

/** @brief Subframe size for samples used for encoded audio data.
 * @details This is the number of bytes in the data stream which
 * which carries a single sample. It is therefore the size of the
 * sample resolution or greater.
 */
#if AUD_BIT_RESOLUTION == 16
#define AUD_SUBFRAME_SIZE 2
#elif AUD_BIT_RESOLUTION == 24
#define AUD_SUBFRAME_SIZE 3
#elif AUD_BIT_RESOLUTION == 32
#define AUD_SUBFRAME_SIZE 2
#else
#error Audio Resolution and Subframe size not allowed.
#endif

/** @brief Filler sample size.
 * @details This is the number of bytes in the data stream which
 * can be used to fill in missing data in the audio stream. It must
 * be a multiple number of 32 bit longwords.
 * 16 bit subframe: 8 bytes is 2 samples
 * 24 bit subframe: 12 bytes is 2 samples
 * 32 bit subframe: 16 bytes is 2 samples
 */
//@{
#define AUD_SAMPLE_FILL_16 8
#define AUD_SAMPLE_FILL_24 12
#define AUD_SAMPLE_FILL_32 16
#if AUD_BIT_RESOLUTION == 16
#define AUD_SAMPLE_FILL AUD_SAMPLE_FILL_16
#elif AUD_BIT_RESOLUTION == 24
#define AUD_SAMPLE_FILL AUD_SAMPLE_FILL_24
#elif AUD_BIT_RESOLUTION == 32
#define AUD_SAMPLE_FILL AUD_SAMPLE_FILL_32
#endif
//@}

/** @brief Payload format definition
 * @details This is the type of the audio samples.
 * Allowed values are USB_AUDIO_FORMAT_TYPE_I or USB_AUDIO_FORMAT_TYPE_II.
 */
#define AUD_FORMAT_TYPE USB_AUDIO_FORMAT_TYPE_I

/** @brief Format used for encoded audio data.
 * @details For 16 bit samples and above, TYPE I PCM should be used.
 * For 8 bit samples then TYPE I PCM8 is used.
 * Allowed values are USB_AUDIO_BMFORMATS_TYPE_I_PCM or
 * USB_AUDIO_BMFORMATS_TYPE_I_PCM8. Note the WM8731 DAC on the
 * MM900EVxA board does not support 8 bit samples.
 */
#define AUD_FORMAT USB_AUDIO_BMFORMATS_TYPE_I_PCM

/** @brief Sample frequency for audio data stream.
 * @details This is the frequency at which the data stream samples
 * are generated. This can normally be 44100 or 48000.
 *
 * Note: MM900EV2 Boards.
 * Due to the MM900EV2 using a 12 MHz crystal for the WM8731 the
 * DAC is run in "USB mode". This means when the sample rate is
 * 44118 Hz we need to add dummy samples to the data stream to
 * compensate. The audio will be pitch shifted (~0.5%) and have
 * artifacts from the added	samples.
 */
//@{
#define AUD_SAMPLE_FREQ 48000
#define AUD_SAMPLE_FREQ_0 ((AUD_SAMPLE_FREQ >> 0) & 0xff) // 0x80
#define AUD_SAMPLE_FREQ_1 ((AUD_SAMPLE_FREQ >> 8) & 0xff)  // 0xbb
#define AUD_SAMPLE_FREQ_2 ((AUD_SAMPLE_FREQ >> 16) & 0xff)  // 0x00
//@}

/** @brief Audio packet length.
 * @details The data stream will consist of USB packets of a certain
 * length depending on the subframe size, number of channels and the
 * sample frequency. The frame pattern for certain sample rates will
 * mean that there may be one extra sample per packet.
 */
//@{
#define AUD_PACKET_LEN_HS (((AUD_SAMPLE_FREQ * AUD_SUBFRAME_SIZE * AUD_NUM_CHANNELS) / \
		(8000 / (1 << (UAD_EP_AUDIO_STREAM_INTERVAL_HS - 1)))) + (AUD_SUBFRAME_SIZE * AUD_NUM_CHANNELS))
#define AUD_PACKET_LEN_FS (((AUD_SAMPLE_FREQ * AUD_SUBFRAME_SIZE * AUD_NUM_CHANNELS) / \
		(1000 / UAD_EP_AUDIO_STREAM_INTERVAL_FS)) + (AUD_SUBFRAME_SIZE * AUD_NUM_CHANNELS))

#if AUD_PACKET_LEN_FS > AUD_PACKET_LEN_HS
#define AUD_PACKET_LEN_MAX AUD_PACKET_LEN_FS
#else
#define AUD_PACKET_LEN_MAX AUD_PACKET_LEN_HS
#endif
//@}

//@}

/**
 @brief Terminal IDs
 */
//@{
#define AUD_ID_INPUT 1  	// Input terminal ID
#define AUD_ID_OUTPUT 2  	// Output terminal ID
//@}

/**
 @brief Number of byte of audio data to buffer.
 */
#define AUDIO_DAC_BUFFER 2048

/* GLOBAL VARIABLES ****************************************************************/

/* LOCAL VARIABLES *****************************************************************/

/** @brief DAC buffer.
 * @details Received data on USB for transmission to I2S bus.
 */
static uint8_t dac_buffer[AUDIO_DAC_BUFFER] __attribute__((aligned(4)));
static uint8_t dac_zero[32] __attribute__((aligned(4)));
static uint8_t recv_packet[AUD_PACKET_LEN_MAX + (AUD_SUBFRAME_SIZE * AUD_NUM_CHANNELS)];

/** @brief DAC write pointer.
 * @details Line position DAC buffer to write data.
 */
static volatile uint16_t dac_wr = 0;

/** @brief DAC read pointer.
 * @details Line position in DAC buffer to read data.
 */
static volatile uint16_t dac_rd = 0;

/** @brief DAC buffer content size.
 * @details Number of waiting bytes in the DAC buffer.
 */
volatile uint32_t dac_buffer_size = 0;

/** @brief DAC state change flag.
 * @details Signals bottom half that a commit has changed something about the dac image.
 */
static uint8_t dac_state_change = 0;

/**
 @name string_descriptor
 @brief Table of USB String descriptors

 This is placed at a fixed location in the const section allowing
 up-to 256 bytes of string descriptors to be defined. These can be
 modified or replaced by a utility or binary editor without
 requiring a recompilation of the firmware.
 They are placed at offset 0x100 and reserve 0x100 bytes.
 The data is not stored in section and is therefore
 regarded as const.
 */
#define UNICODE_LEN(A) (((A * 2) + 2) | (USB_DESCRIPTOR_TYPE_STRING << 8))

const uint16_t string_descriptor[STRING_DESCRIPTOR_ALLOCATION/sizeof(uint16_t)] =
{
		UNICODE_LEN(1), 0x0409, // 0409 = English (US)
		// String 1 (Manufacturer): "FTDI"
		UNICODE_LEN(4), L'F', L'T', L'D', L'I',
		// String 2 (Product): Depends on settings
#if AUD_FORMAT_TYPE == USB_AUDIO_FORMAT_TYPE_I
#if AUD_FORMAT == USB_AUDIO_BMFORMATS_TYPE_I_PCM
		UNICODE_LEN(32),
			L'A', L'N', L'_', L'4', L'3', L'3', L' ',
			L'T', L'y', L'p', L'e', L'I', L' ',
			L'P', L'C', L'M', L' ',
#if AUD_BIT_RESOLUTION == 16
			L'1', L'6', L' ', L'b', L'i', L't', L' ',
#elif AUD_BIT_RESOLUTION == 24
			L'2', L'4', L' ', L'b', L'i', L't', L' ',
#elif AUD_BIT_RESOLUTION == 32
			L'3', L'2', L' ', L'b', L'i', L't', L' ',
#endif
#if AUD_SAMPLE_FREQ == 48000
			L'4', L'8', L'0', L'0', L'0', L' ', L'H', L'z',
#elif AUD_SAMPLE_FREQ == 44100
			L'4', L'4', L'1', L'0', L'0', L' ', L'H', L'z',
#endif
#elif AUD_FORMAT == USB_AUDIO_BMFORMATS_TYPE_I_PCM8
		UNICODE_LEN(33), L'A', L'N', L'_', L'4', L'3', L'3', L' ',
			L'T', L'y', L'p', L'e', L'I', L' ',
			L'P', L'C', L'M', L'8', L' ',
#if AUD_BIT_RESOLUTION == 16
			L'1', L'6', L' ', L'b', L'i', L't', L' ',
#elif AUD_BIT_RESOLUTION == 24
			L'2', L'4', L' ', L'b', L'i', L't', L' ',
#elif AUD_BIT_RESOLUTION == 32
			L'3', L'2', L' ', L'b', L'i', L't', L' ',
#endif
#if AUD_SAMPLE_FREQ == 48000
			L'4', L'8', L'0', L'0', L'0', L' ', L'H', L'z',
#elif AUD_SAMPLE_FREQ == 44100
			L'4', L'4', L'1', L'0', L'0', L' ', L'H', L'z',
#endif
#endif

#elif AUD_FORMAT_TYPE == USB_AUDIO_FORMAT_TYPE_II
		UNICODE_LEN(33), L'A', L'N', L'_', L'4', L'3', L'3', L' ',
			L'T', L'y', L'p', L'e', L'I', L'I', L' ',
			L'P', L'C', L'M', L' ',
#if AUD_BIT_RESOLUTION == 16
			L'1', L'6', L' ', L'b', L'i', L't', L' ',
#elif AUD_BIT_RESOLUTION == 24
			L'2', L'4', L' ', L'b', L'i', L't', L' ',
#elif AUD_BIT_RESOLUTION == 32
			L'3', L'2', L' ', L'b', L'i', L't', L' ',
#endif
#if AUD_SAMPLE_FREQ == 48000
			L'4', L'8', L'0', L'0', L'0', L' ', L'H', L'z',
#elif AUD_SAMPLE_FREQ == 44100
			L'4', L'4', L'1', L'0', L'0', L' ', L'H', L'z',
#endif
#endif
		// String 3 (Serial Number):
		UNICODE_LEN(8), L'F', L'T', L'0', L'0', L'0', L'0', L'0', L'0',
		// String 4 (DFU Product Name): "FT900 DFU Mode"
		UNICODE_LEN(14), L'F', L'T', L'9', L'0', L'0', L' ', L'D', L'F', L'U', L' ', L'M', L'o', L'd', L'e',
		// String 5 (Interface Name): "DFU Interface"
		UNICODE_LEN(13), L'D', L'F', L'U', L' ', L'I', L'n', L't', L'e', L'r', L'f', L'a', L'c', L'e',
		// END OF STRINGS
		0x0000,
};
//@}

/**
 @name wcid_string_descriptor
 @brief USB String descriptor for WCID identification.
 */
DESCRIPTOR_QUALIFIER uint8_t wcid_string_descriptor[USB_MICROSOFT_WCID_STRING_LENGTH] = {
		USB_MICROSOFT_WCID_STRING(WCID_VENDOR_REQUEST_CODE)
};

/**
 @name device_descriptor_uvc
 @brief Device descriptor for Run Time mode.
 */
DESCRIPTOR_QUALIFIER USB_device_descriptor device_descriptor_uvc =
{
		sizeof(USB_device_descriptor), /* bLength */
		USB_DESCRIPTOR_TYPE_DEVICE, /* bDescriptorType */
		USB_BCD_VERSION_2_0, /* bcdUSB */          // V2.0
		USB_CLASS_DEVICE, /* bDeviceClass */       // Defined in interface
		USB_SUBCLASS_DEVICE, /* bDeviceSubClass */ // Defined in interface
		USB_PROTOCOL_DEVICE, /* bDeviceProtocol */ // Defined in interface
		USB_CONTROL_EP_MAX_PACKET_SIZE, /* bMaxPacketSize0 */
		USB_VID_FTDI, /* idVendor */   // idVendor: 0x0403 (FTDI)
		USB_PID_AUDIO, /* idProduct */ // idProduct: 0x0fd5
		0x0101, /* bcdDevice */        // 1.1
		0x01, /* iManufacturer */      // Manufacturer
		0x02, /* iProduct */           // Product
		0x03, /* iSerialNumber */      // Serial Number
		0x01, /* bNumConfigurations */
};

/**
 @name device_qualifier_descriptor_uvc
 @brief Device qualifier descriptor for Run Time mode.
 */
DESCRIPTOR_QUALIFIER USB_device_qualifier_descriptor device_qualifier_descriptor_uvc =
{
		sizeof(USB_device_qualifier_descriptor), /* bLength */
		USB_DESCRIPTOR_TYPE_DEVICE_QUALIFIER, /* bDescriptorType */
		USB_BCD_VERSION_2_0, /* bcdUSB */          // V2.0
		USB_CLASS_MISCELLANEOUS, /* bDeviceClass */       // Defined in interface
		USB_SUBCLASS_COMMON_CLASS, /* bDeviceSubClass */ // Defined in interface
		USB_PROTOCOL_INTERFACE_ASSOCIATION, /* bDeviceProtocol */ // Defined in interface
		USB_CONTROL_EP_MAX_PACKET_SIZE, /* bMaxPacketSize0 */
		1, /* bNumConfigurations */
		0
};


/**
 @brief Configuration descriptor for Run Time mode.
 */
//@{
/**
 @brief AudioControl Configuration descriptor for Run Time mode.
 */
struct PACK UAD_AC_config_descriptor {
	usb_audioControlInterfaceHeaderDescriptor_t(1) ac_header;
	usb_audioInputTerminalDescriptor_t input_terminal;
	usb_audioOutputTerminalDescriptor_t output_terminal;
};

/**
 @brief AudioStreaming Configuration descriptor for High Speed.
 */
struct PACK UVC_AS_config_descriptor {
	usb_audioStreamingClassInterfaceDescriptor_t as_class;
	usb_audioFormatTypeIDescriptor_t(1) as_typeI;
	usb_audioStreamingDataStandardEndpointDescriptor_t ep_data_out_1;
	usb_audioStreamingDataClassEndpointDescriptor_t as_ac_ep;
};


/**
 @brief Configuration descriptor - High Speed.
 */
struct PACK config_descriptor_uad_hs
{
	USB_configuration_descriptor configuration;

	usb_audioControlInterfaceDescriptor_t interface_audio_control;
	struct UAD_AC_config_descriptor ac_config;

	usb_audioStreamingStandardInterfaceDescriptor_t interface_audio_stream_alt0;
	usb_audioStreamingStandardInterfaceDescriptor_t interface_audio_stream_alt1;
	struct UVC_AS_config_descriptor as_config;

#ifdef USB_INTERFACE_USE_DFU
	USB_interface_descriptor dfu_interface;
	USB_dfu_functional_descriptor dfu_functional;
#endif // USB_INTERFACE_USE_DFU
};

/**
 @brief Configuration descriptor - Full Speed.
 */
struct PACK config_descriptor_uad_fs
{
	USB_configuration_descriptor configuration;

	usb_audioControlInterfaceDescriptor_t interface_audio_control;
	struct UAD_AC_config_descriptor ac_config;

	usb_audioStreamingStandardInterfaceDescriptor_t interface_audio_stream_alt0;
	usb_audioStreamingStandardInterfaceDescriptor_t interface_audio_stream_alt1;
	struct UVC_AS_config_descriptor as_config;

#ifdef USB_INTERFACE_USE_DFU
	USB_interface_descriptor dfu_interface;
	USB_dfu_functional_descriptor dfu_functional;
#endif // USB_INTERFACE_USE_DFU
};

/**
 @brief Configuration descriptor declaration and initialisation.
 */
DESCRIPTOR_QUALIFIER struct config_descriptor_uad_hs config_descriptor_uad_hs =
{
		{
				sizeof(USB_configuration_descriptor), /* configuration.bLength */
				USB_DESCRIPTOR_TYPE_CONFIGURATION, /* configuration.bDescriptorType */
				sizeof(struct config_descriptor_uad_hs), /* configuration.wTotalLength */
#ifdef USB_INTERFACE_USE_DFU
				0x03, /* configuration.bNumInterfaces */
#else // !USB_INTERFACE_USE_DFU
				0x02, /* configuration.bNumInterfaces */
#endif // USB_INTERFACE_USE_DFU
				0x01, /* configuration.bConfigurationValue */
				0x00, /* configuration.iConfiguration */
				USB_CONFIG_BMATTRIBUTES_VALUE, /* configuration.bmAttributes */
				0xFA, /* configuration.bMaxPower */           // 500mA
		},

		{
				sizeof(usb_audioControlInterfaceDescriptor_t), /*interface_audio_control.bLength */
				USB_DESCRIPTOR_TYPE_INTERFACE, /*interface_audio_control.bDescriptorType */
				0, /*interface_audio_control.bInterfaceNumber */
				0, /*interface_audio_control.bAlternateSetting */
				0, /*interface_audio_control.bNumEndpoints */
				USB_CLASS_AUDIO, /*interface_audio_control.bInterfaceClass */
				USB_SUBCLASS_AUDIO_AUDIOCONTROL, /*interface_audio_control.bInterfaceSubClass */
				INTERFACE_PROTOCOL_UNDEFINED, /*interface_audio_control.bInterfaceProtocol */
				0, /*interface_audio_control.iInterface */
		},

		{
				{
						sizeof(usb_audioControlInterfaceHeaderDescriptor_t(1)), /* ac_header.bLength */
						USB_DESCRIPTOR_TYPE_AUDIO_INTERFACE, /* ac_header.bDescriptorType */
						USB_DESCRIPTOR_SUBTYPE_AUDIO_CONTROL_HEADER, /* ac_header.bDescriptorSubtype */
						0x0100, /* ac_header.bcdADC */
						sizeof(struct UAD_AC_config_descriptor), /* ac_header.wTotalLength */
						1, /* ac_header.bInCollection */
						{ 1 }, /* ac_header.baInterfaceNr */
				},

				{
						sizeof(usb_audioInputTerminalDescriptor_t), /* input_terminal.bLength */
						USB_DESCRIPTOR_TYPE_AUDIO_INTERFACE, /* input_terminal.bDescriptorType */
						USB_DESCRIPTOR_SUBTYPE_AUDIO_CONTROL_INPUT_TERMINAL, /* input_terminal.bDescriptorSubtype */
						AUD_ID_INPUT, /* input_terminal.bTerminalID */
						USB_AUDIO_TERMINAL_TYPE_STREAMING, /* input_terminal.wTerminalType */
						0, /* input_terminal.bAssocTerminal */
						AUD_NUM_CHANNELS, /* input_terminal.bNrChannels */
						AUD_CHANNEL_CONFIG, /* input_terminal.bmChannelConfig */
						0, /* input_terminal.iChannelNames */
						0, /* input_terminal.iTerminal */
				},

				{
						sizeof(usb_audioOutputTerminalDescriptor_t), /* output_terminal.bLength */
						USB_DESCRIPTOR_TYPE_AUDIO_INTERFACE, /* output_terminal.bDescriptorType */
						USB_DESCRIPTOR_SUBTYPE_AUDIO_CONTROL_OUTPUT_TERMINAL, /* output_terminal.bDescriptorSubtype */
						AUD_ID_OUTPUT, /* output_terminal.bTerminalID */
						USB_AUDIO_TERMINAL_TYPE_OUTPUT_SPEAKER, /* output_terminal.wTerminalType */
						0, /* output_terminal.bAssocTerminal */
						AUD_ID_INPUT, /* output_terminal.bSourceID */
						0, /* output_terminal.iTerminal */
				},
		},

		// Alt setting 0 - no isochronous endpoints
		{
				sizeof(usb_audioStreamingStandardInterfaceDescriptor_t), /* interface_audio_stream.bLength */
				USB_DESCRIPTOR_TYPE_INTERFACE, /* interface_audio_stream.bDescriptorType */
				1, /* interface_audio_stream.bInterfaceNumber */
				0, /* interface_audio_stream.bAlternateSetting */
				0, /* interface_audio_stream.bNumEndpoints */
				USB_CLASS_AUDIO, /* interface_audio_stream.bInterfaceClass */
				USB_SUBCLASS_AUDIO_AUDIOSTREAMING, /* interface_audio_stream.bInterfaceSubClass */
				INTERFACE_PROTOCOL_UNDEFINED, /* interface_audio_stream.bInterfaceProtocol */
				0, /* interface_audio_stream.iInterface */
		},

		// Alt setting 1 - 2 isochronous endpoints
		{
				sizeof(usb_audioStreamingStandardInterfaceDescriptor_t), /* interface_audio_stream.bLength */
				USB_DESCRIPTOR_TYPE_INTERFACE, /* interface_audio_stream.bDescriptorType */
				1, /* interface_audio_stream.bInterfaceNumber */
				1, /* interface_audio_stream.bAlternateSetting */
				1, /* interface_audio_stream.bNumEndpoints */
				USB_CLASS_AUDIO, /* interface_audio_stream.bInterfaceClass */
				USB_SUBCLASS_AUDIO_AUDIOSTREAMING, /* interface_audio_stream.bInterfaceSubClass */
				INTERFACE_PROTOCOL_UNDEFINED, /* interface_audio_stream.bInterfaceProtocol */
				0, /* interface_audio_stream.iInterface */
		},

		{
				{
					    sizeof(usb_audioStreamingClassInterfaceDescriptor_t), /* as_class.bLength */
						USB_DESCRIPTOR_TYPE_AUDIO_INTERFACE, /* as_class.bDescriptorType */
						USB_DESCRIPTOR_SUBTYPE_AUDIO_STREAMING_GENERAL, /* as_class.bDescriptorSubtype */
						AUD_ID_INPUT, /* as_class.bTerminalLink */
						1, /* as_class.bDelay */
						AUD_FORMAT, /* as_class.bmFormats */
				},

#if AUD_FORMAT_TYPE == USB_AUDIO_FORMAT_TYPE_I
				{
					    sizeof(usb_audioFormatTypeIDescriptor_t(1)), /* as_typeI.bLength */
						USB_DESCRIPTOR_TYPE_AUDIO_INTERFACE, /* as_typeI.bDescriptorType */
						USB_DESCRIPTOR_SUBTYPE_AUDIO_STREAMING_FORMAT_TYPE, /* as_typeI.bDescriptorSubtype */
						USB_AUDIO_FORMAT_TYPE_I, /* as_typeI.bFormatType */
						AUD_NUM_CHANNELS, /* as_typeI.bNrChannels */
						AUD_SUBFRAME_SIZE, /* as_typeI.bSubFrameSize */
						AUD_BIT_RESOLUTION, /* as_typeI.bBitResolution */
						1, /* as_typeI.bSamFreqType */
						{
							{ AUD_SAMPLE_FREQ_0, AUD_SAMPLE_FREQ_1, AUD_SAMPLE_FREQ_2, } /* as_typeI.tSamFreq */
						},
				},

#else // AUD_FORMAT_TYPE == USB_AUDIO_FORMAT_TYPE_II

#endif // AUD_FORMAT_TYPE
				{
						sizeof(usb_audioStreamingDataStandardEndpointDescriptor_t), /* ep_data_out_1.bLength */
						USB_DESCRIPTOR_TYPE_ENDPOINT, /* ep_data_out_1.bDescriptorType */
						UAD_EP_AUDIO_STREAM, /* ep_data_out_1.bEndpointAddress */
						USB_ENDPOINT_DESCRIPTOR_ATTR_ISOCHRONOUS | USB_ENDPOINT_DESCRIPTOR_ISOCHRONOUS_ASYNCHRONOUS, /* ep_data_out_1.bmAttributes */
						AUD_PACKET_LEN_HS, // UAD_EP_AUDIO_STREAM_SIZE, /* ep_data_out_1.wMaxPacketSize */
						UAD_EP_AUDIO_STREAM_INTERVAL_HS, /* ep_data_out_1.bInterval */
						0, /* ep_data_out_1.bRefresh */
						0, /* ep_data_out_1.bSynchAddress */
				},

				{
					    sizeof(usb_audioStreamingDataClassEndpointDescriptor_t), /* as_ac_ep.bLength */
						USB_DESCRIPTOR_TYPE_AUDIO_ENDPOINT, /* as_ac_ep.bDescriptorType */
						USB_ENDPOINT_DESCRIPTOR_AUDIO_GENERAL, /* as_ac_ep.bDescriptorSubtype */
						USB_AUDIO_ENDPOINT_BMATTRIBUTE_MAX_PACKET, /* as_ac_ep.bmAttributes */
						0, /* as_ac_ep.bLockDelayUnits */
						0, /* as_ac_ep.wLockDelay */
				},
		},

#ifdef USB_INTERFACE_USE_DFU
		// ---- INTERFACE DESCRIPTOR for DFU Interface ----
		{
				sizeof(USB_interface_descriptor), /* dfu_interface.bLength */
				USB_DESCRIPTOR_TYPE_INTERFACE, /* dfu_interface.bDescriptorType */
				DFU_USB_INTERFACE_RUNTIME, /* dfu_interface.bInterfaceNumber */
				0x00, /* dfu_interface.bAlternateSetting */
				0x00, /* dfu_interface.bNumEndpoints */
				USB_CLASS_APPLICATION, /* dfu_interface.bInterfaceClass */ // bInterfaceClass: Application Specific Class
				USB_SUBCLASS_DFU, /* dfu_interface.bInterfaceSubClass */ // bInterfaceSubClass: Device Firmware Update
				USB_PROTOCOL_DFU_RUNTIME, /* dfu_interface.bInterfaceProtocol */ // bInterfaceProtocol: Runtime Protocol
				0x05 /* dfu_interface.iInterface */       // * iInterface: "DFU Interface"
		},

		// ---- FUNCTIONAL DESCRIPTOR for DFU Interface ----
		{
				sizeof(USB_dfu_functional_descriptor), /* dfu_functional.bLength */
				USB_DESCRIPTOR_TYPE_DFU_FUNCTIONAL, /* dfu_functional.bDescriptorType */
				DFU_ATTRIBUTES, /* dfu_functional.bmAttributes */  	// bmAttributes
				DFU_TIMEOUT, /* dfu_functional.wDetatchTimeOut */ // wDetatchTimeOut
				DFU_TRANSFER_SIZE, /* dfu_functional.wTransferSize */     // wTransferSize
				USB_BCD_VERSION_DFU_1_1 /* dfu_functional.bcdDfuVersion */ // bcdDfuVersion: DFU Version 1.1
		}
#endif // USB_INTERFACE_USE_DFU
};

DESCRIPTOR_QUALIFIER struct config_descriptor_uad_fs config_descriptor_uad_fs =
{
		{
				sizeof(USB_configuration_descriptor), /* configuration.bLength */
				USB_DESCRIPTOR_TYPE_CONFIGURATION, /* configuration.bDescriptorType */
				sizeof(struct config_descriptor_uad_fs), /* configuration.wTotalLength */
				0x02, /* configuration.bNumInterfaces */
				0x01, /* configuration.bConfigurationValue */
				0x00, /* configuration.iConfiguration */
				USB_CONFIG_BMATTRIBUTES_VALUE, /* configuration.bmAttributes */
				0xFA, /* configuration.bMaxPower */           // 500mA
		},

		{
				sizeof(usb_audioControlInterfaceDescriptor_t), /*interface_audio_control.bLength */
				USB_DESCRIPTOR_TYPE_INTERFACE, /*interface_audio_control.bDescriptorType */
				0, /*interface_audio_control.bInterfaceNumber */
				0, /*interface_audio_control.bAlternateSetting */
				0, /*interface_audio_control.bNumEndpoints */
				USB_CLASS_AUDIO, /*interface_audio_control.bInterfaceClass */
				USB_SUBCLASS_AUDIO_AUDIOCONTROL, /*interface_audio_control.bInterfaceSubClass */
				INTERFACE_PROTOCOL_UNDEFINED, /*interface_audio_control.bInterfaceProtocol */
				0, /*interface_audio_control.iInterface */
		},

		{
				{
						sizeof(usb_audioControlInterfaceHeaderDescriptor_t(1)), /* ac_header.bLength */
						USB_DESCRIPTOR_TYPE_AUDIO_INTERFACE, /* ac_header.bDescriptorType */
						USB_DESCRIPTOR_SUBTYPE_AUDIO_CONTROL_HEADER, /* ac_header.bDescriptorSubtype */
						0x0100, /* ac_header.bcdADC */
						sizeof(struct UAD_AC_config_descriptor), /* ac_header.wTotalLength */
						1, /* ac_header.bInCollection */
						{ 1 }, /* ac_header.baInterfaceNr */
				},

				{
						sizeof(usb_audioInputTerminalDescriptor_t), /* input_terminal.bLength */
						USB_DESCRIPTOR_TYPE_AUDIO_INTERFACE, /* input_terminal.bDescriptorType */
						USB_DESCRIPTOR_SUBTYPE_AUDIO_CONTROL_INPUT_TERMINAL, /* input_terminal.bDescriptorSubtype */
						AUD_ID_INPUT, /* input_terminal.bTerminalID */
						USB_AUDIO_TERMINAL_TYPE_STREAMING, /* input_terminal.wTerminalType */
						0, /* input_terminal.bAssocTerminal */
						AUD_NUM_CHANNELS, /* input_terminal.bNrChannels */
						/* input_terminal.bmChannelConfig */
						AUD_CHANNEL_CONFIG,
						0, /* input_terminal.iChannelNames */
						0, /* input_terminal.iTerminal */
				},

				{
						sizeof(usb_audioOutputTerminalDescriptor_t), /* output_terminal.bLength */
						USB_DESCRIPTOR_TYPE_AUDIO_INTERFACE, /* output_terminal.bDescriptorType */
						USB_DESCRIPTOR_SUBTYPE_AUDIO_CONTROL_OUTPUT_TERMINAL, /* output_terminal.bDescriptorSubtype */
						AUD_ID_OUTPUT, /* output_terminal.bTerminalID */
						USB_AUDIO_TERMINAL_TYPE_OUTPUT_SPEAKER, /* output_terminal.wTerminalType */
						0, /* output_terminal.bAssocTerminal */
						AUD_ID_INPUT, /* output_terminal.bSourceID */
						0, /* output_terminal.iTerminal */
				},
		},

		// Alt setting 0 - no isochronous endpoints
		{
				sizeof(usb_audioStreamingStandardInterfaceDescriptor_t), /* interface_audio_stream.bLength */
				USB_DESCRIPTOR_TYPE_INTERFACE, /* interface_audio_stream.bDescriptorType */
				1, /* interface_audio_stream.bInterfaceNumber */
				0, /* interface_audio_stream.bAlternateSetting */
				0, /* interface_audio_stream.bNumEndpoints */
				USB_CLASS_AUDIO, /* interface_audio_stream.bInterfaceClass */
				USB_SUBCLASS_AUDIO_AUDIOSTREAMING, /* interface_audio_stream.bInterfaceSubClass */
				INTERFACE_PROTOCOL_UNDEFINED, /* interface_audio_stream.bInterfaceProtocol */
				0, /* interface_audio_stream.iInterface */
		},

		// Alt setting 1 - 2 isochronous endpoints
		{
				sizeof(usb_audioStreamingStandardInterfaceDescriptor_t), /* interface_audio_stream.bLength */
				USB_DESCRIPTOR_TYPE_INTERFACE, /* interface_audio_stream.bDescriptorType */
				1, /* interface_audio_stream.bInterfaceNumber */
				1, /* interface_audio_stream.bAlternateSetting */
				2, /* interface_audio_stream.bNumEndpoints */
				USB_CLASS_AUDIO, /* interface_audio_stream.bInterfaceClass */
				USB_SUBCLASS_AUDIO_AUDIOSTREAMING, /* interface_audio_stream.bInterfaceSubClass */
				INTERFACE_PROTOCOL_UNDEFINED, /* interface_audio_stream.bInterfaceProtocol */
				0, /* interface_audio_stream.iInterface */
		},

		{
				{
					    sizeof(usb_audioStreamingClassInterfaceDescriptor_t), /* as_class.bLength */
						USB_DESCRIPTOR_TYPE_AUDIO_INTERFACE, /* as_class.bDescriptorType */
						USB_DESCRIPTOR_SUBTYPE_AUDIO_STREAMING_GENERAL, /* as_class.bDescriptorSubtype */
						AUD_ID_INPUT, /* as_class.bTerminalLink */
						1, /* as_class.bDelay */
						AUD_FORMAT, /* as_class.bmFormats */
				},

#if AUD_FORMAT_TYPE == USB_AUDIO_FORMAT_TYPE_I
				{
					    sizeof(usb_audioFormatTypeIDescriptor_t(1)), /* as_typeI.bLength */
						USB_DESCRIPTOR_TYPE_AUDIO_INTERFACE, /* as_typeI.bDescriptorType */
						USB_DESCRIPTOR_SUBTYPE_AUDIO_STREAMING_FORMAT_TYPE, /* as_typeI.bDescriptorSubtype */
						USB_AUDIO_FORMAT_TYPE_I, /* as_typeI.bFormatType */
						AUD_NUM_CHANNELS, /* as_typeI.bNrChannels */
						AUD_SUBFRAME_SIZE, /* as_typeI.bSubFrameSize */
						AUD_BIT_RESOLUTION, /* as_typeI.bBitResolution */
						1, /* as_typeI.bSamFreqType */
						{
							{ AUD_SAMPLE_FREQ_0, AUD_SAMPLE_FREQ_1, AUD_SAMPLE_FREQ_2, }, /* as_typeI.tSamFreq */
						},
				},

#else

#endif // AUD_FORMAT_TYPE
				{
						sizeof(usb_audioStreamingDataStandardEndpointDescriptor_t), /* ep_data_out_1.bLength */
						USB_DESCRIPTOR_TYPE_ENDPOINT, /* ep_data_out_1.bDescriptorType */
						UAD_EP_AUDIO_STREAM, /* ep_data_out_1.bEndpointAddress */
						USB_ENDPOINT_DESCRIPTOR_ATTR_ISOCHRONOUS | USB_ENDPOINT_DESCRIPTOR_ISOCHRONOUS_ASYNCHRONOUS, /* ep_data_out_1.bmAttributes */
						AUD_PACKET_LEN_FS, //UAD_EP_AUDIO_STREAM_SIZE, /* ep_data_out_1.wMaxPacketSize */
						UAD_EP_AUDIO_STREAM_INTERVAL_FS, /* ep_data_out_1.bInterval */
						0, /* ep_data_out_1.bRefresh */
						0, /* ep_data_out_1.bSynchAddress */
				},

				{
					    sizeof(usb_audioStreamingDataClassEndpointDescriptor_t), /* as_ac_ep.bLength */
						USB_DESCRIPTOR_TYPE_AUDIO_ENDPOINT, /* as_ac_ep.bDescriptorType */
						USB_ENDPOINT_DESCRIPTOR_AUDIO_GENERAL, /* as_ac_ep.bDescriptorSubtype */
						USB_AUDIO_ENDPOINT_BMATTRIBUTE_MAX_PACKET, /* as_ac_ep.bmAttributes */
						0, /* as_ac_ep.bLockDelayUnits */
						0, /* as_ac_ep.wLockDelay */
				},
		},

#ifdef USB_INTERFACE_USE_DFU
		// ---- INTERFACE DESCRIPTOR for DFU Interface ----
		{
				sizeof(USB_interface_descriptor), /* dfu_interface.bLength */
				USB_DESCRIPTOR_TYPE_INTERFACE, /* dfu_interface.bDescriptorType */
				DFU_USB_INTERFACE_RUNTIME, /* dfu_interface.bInterfaceNumber */
				0x00, /* dfu_interface.bAlternateSetting */
				0x00, /* dfu_interface.bNumEndpoints */
				USB_CLASS_APPLICATION, /* dfu_interface.bInterfaceClass */ // bInterfaceClass: Application Specific Class
				USB_SUBCLASS_DFU, /* dfu_interface.bInterfaceSubClass */ // bInterfaceSubClass: Device Firmware Update
				USB_PROTOCOL_DFU_RUNTIME, /* dfu_interface.bInterfaceProtocol */ // bInterfaceProtocol: Runtime Protocol
				0x05 /* dfu_interface.iInterface */       // * iInterface: "DFU Interface"
		},

		// ---- FUNCTIONAL DESCRIPTOR for DFU Interface ----
		{
				sizeof(USB_dfu_functional_descriptor), /* dfu_functional.bLength */
				USB_DESCRIPTOR_TYPE_DFU_FUNCTIONAL, /* dfu_functional.bDescriptorType */
				DFU_ATTRIBUTES, /* dfu_functional.bmAttributes */  	// bmAttributes
				DFU_TIMEOUT, /* dfu_functional.wDetatchTimeOut */ // wDetatchTimeOut
				DFU_TRANSFER_SIZE, /* dfu_functional.wTransferSize */     // wTransferSize
				USB_BCD_VERSION_DFU_1_1 /* dfu_functional.bcdDfuVersion */ // bcdDfuVersion: DFU Version 1.1
		}
#endif // USB_INTERFACE_USE_DFU
};
//@}

#ifdef USB_INTERFACE_USE_DFU
/**
 @name wcid_feature_runtime
 @brief WCID Compatible ID for DFU interface in runtime.
 */
//@{
DESCRIPTOR_QUALIFIER USB_WCID_feature_descriptor wcid_feature_runtime =
{
		sizeof(struct USB_WCID_feature_descriptor), /* dwLength */
		USB_MICROSOFT_WCID_VERSION, /* bcdVersion */
		USB_MICROSOFT_WCID_FEATURE_WINDEX_COMPAT_ID, /* wIndex */
		1, /* bCount */
		{0, 0, 0, 0, 0, 0, 0,}, /* rsv1 */
		DFU_USB_INTERFACE_RUNTIME, /* bFirstInterfaceNumber */
		1, /* rsv2 - set to 1 */
		{'W', 'I', 'N', 'U', 'S', 'B', 0x00, 0x00,}, /* compatibleID[8] */
		{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,}, /* subCompatibleID[8] */
		{0, 0, 0, 0, 0, 0,}, /* rsv3[6] */
};
#endif // USB_INTERFACE_USE_DFU

/**
 @name device_descriptor_dfumode
 @brief Device descriptor for DFU Mode.
 */
DESCRIPTOR_QUALIFIER USB_device_descriptor device_descriptor_dfumode =
{
		0x12, /* bLength */
		USB_DESCRIPTOR_TYPE_DEVICE, /* bDescriptorType */
		USB_BCD_VERSION_2_0, /* bcdUSB */          // V2.00
		USB_CLASS_DEVICE, /* bDeviceClass */       // Defined in interface
		USB_SUBCLASS_DEVICE, /* bDeviceSubClass */ // Defined in interface
		USB_PROTOCOL_DEVICE, /* bDeviceProtocol */ // Defined in interface
		USB_CONTROL_EP_MAX_PACKET_SIZE, /* bMaxPacketSize0 */    // 8
		USB_VID_FTDI, /* idVendor */   // idVendor: 0x0403 (FTDI)
		DFU_USB_PID_DFUMODE, /* idProduct */ // idProduct: 0x0fee
		0x0101, /* bcdDevice */        // 1.1
		0x01, /* iManufacturer */      // Manufacturer
		0x04, /* iProduct */           // Product
		0x03, /* iSerialNumber */      // Serial Number
		0x01, /* bNumConfigurations */
};

/**
 @name config_descriptor_dfumode
 @brief Config descriptor for DFU Mode.
 */
//@{
struct PACK config_descriptor_dfumode
{
	USB_configuration_descriptor configuration;
	USB_interface_descriptor dfu_interface;
	USB_dfu_functional_descriptor dfu_functional;
};

DESCRIPTOR_QUALIFIER struct config_descriptor_dfumode config_descriptor_dfumode =
{
		{
				0x09, /* configuration.bLength */
				USB_DESCRIPTOR_TYPE_CONFIGURATION, /* configuration.bDescriptorType */
				sizeof(struct config_descriptor_dfumode), /* configuration.wTotalLength */
				0x01, /* configuration.bNumInterfaces */
				0x01, /* configuration.bConfigurationValue */
				0x00, /* configuration.iConfiguration */
				USB_CONFIG_BMATTRIBUTES_VALUE, /* configuration.bmAttributes */
				0xFA /* configuration.bMaxPower */ // 500mA
		},

		// ---- INTERFACE DESCRIPTOR for DFU Interface ----
		{
				0x09, /* dfu_interface.bLength */
				USB_DESCRIPTOR_TYPE_INTERFACE, /* dfu_interface.bDescriptorType */
				DFU_USB_INTERFACE_DFUMODE, /* dfu_interface.bInterfaceNumber */
				0x00, /* dfu_interface.bAlternateSetting */
				0x00, /* dfu_interface.bNumEndpoints */
				USB_CLASS_APPLICATION, /* dfu_interface.bInterfaceClass */ // bInterfaceClass: Application Specific Class
				USB_SUBCLASS_DFU, /* dfu_interface.bInterfaceSubClass */ // bInterfaceSubClass: Device Firmware Update
				USB_PROTOCOL_DFU_DFUMODE, /* dfu_interface.bInterfaceProtocol */ // bInterfaceProtocol: Runtime Protocol
				0x05 /* dfu_interface.iInterface */
		},

		// ---- FUNCTIONAL DESCRIPTOR for DFU Interface ----
		{
				0x09, /* dfu_functional.bLength */
				USB_DESCRIPTOR_TYPE_DFU_FUNCTIONAL, /* dfu_functional.bDescriptorType */
				DFU_ATTRIBUTES, /* dfu_functional.bmAttributes */  	// bmAttributes
				DFU_TIMEOUT, /* dfu_functional.wDetatchTimeOut */ // wDetatchTimeOut
				DFU_TRANSFER_SIZE, /* dfu_functional.wTransferSize */     // wTransferSize
				USB_BCD_VERSION_DFU_1_1 /* dfu_functional.bcdDfuVersion */ // bcdDfuVersion: DFU Version 1.1
		}
};
//@}

/**
 @name wcid_feature_dfumode
 @brief WCID Compatible ID for DFU interface in DFU mode.
 */
//@{
DESCRIPTOR_QUALIFIER USB_WCID_feature_descriptor wcid_feature_dfumode =
{
		sizeof(struct USB_WCID_feature_descriptor), /* dwLength */
		USB_MICROSOFT_WCID_VERSION, /* bcdVersion */
		USB_MICROSOFT_WCID_FEATURE_WINDEX_COMPAT_ID, /* wIndex */
		1, /* bCount */
		{0, 0, 0, 0, 0, 0, 0,}, /* rsv1 */
		DFU_USB_INTERFACE_DFUMODE, /* bFirstInterfaceNumber */
		1, /* rsv2 - set to 1 */
		{'W', 'I', 'N', 'U', 'S', 'B', 0x00, 0x00,}, /* compatibleID[8] */
		{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,}, /* subCompatibleID[8] */
		{0, 0, 0, 0, 0, 0,}, /* rsv3[6] */
};

/* GLOBAL VARIABLES ****************************************************************/

/* LOCAL VARIABLES *****************************************************************/

/**
 @brief Millisecond counter
 @details Count-up timer to provide the elapsed time for network operations.
 */
static uint32_t milliseconds = 0;

/**
 @brief Active Alternate Setting
 @details Current active alternate setting for the USB interface.
 */
static uint8_t usb_alt = 0;

/**
 @brief Storage for Configuration Descriptors.
 @details Configuration descriptors may need to be modified to turn from type
 USB_DESCRIPTOR_TYPE_CONFIGURATION to USB_DESCRIPTOR_TYPE_OTHER_SPEED_CONFIGURATION.
 */
union config_descriptor_buffer {
	struct config_descriptor_uad_hs hs;
	struct config_descriptor_uad_fs fs;
} config_descriptor_buffer;

/**
 @brief I2S Setup
 */
//@{
typedef struct
{
    uint8_t addr;
    uint8_t reg;
} reg_t;

#if I2S_TYPE == WM8731

/** @brief WM8741 Registers
*/
//@{
enum WM8731_reg {
	/**	0x2 Left Headphone Out -
			   bit 6:0 = 1111001 : LHPVOL[6:0] - 1111001 = Output Volume to 0dB
			   bit 7   = 1       : LZCEN       - Left Channel Zero Cross detect 1 = Enable
	*/
	WM8731_HPVOLL = (0x02 << 1),
	/**	0x3 Right Headphone Out -
			   bit 6:0 = 1111001 : RHPVOL[6:0] - 1111001 = Output Volume to 0dB
			   bit 7   = 1       : RZCEN       - Right Channel Zero Cross detect 1 = Enable
	*/
	WM8731_HPVOLR = (0x03 << 1),
	/**	0x4 Analogue Audio Path Control -
			   bit 0 = 1 : MICBOOST     - 1 = Enable Boost
			   bit 1 = 1 : MUTEMIC      - 1 = Enable Mute
			   bit 2 = 0 : INSEL        - 0 = Microphone Input Select to ADC
			   bit 3 = 0 : BYPASS       - 0 = Disable Bypass
			   bit 4 = 1 : DACSEL       - 1 = Select DAC
			   bit 5 = 0 : SIDETONE     - 0 = Disable Side Tone Attenuation
			   bit 6 = 0 : SIDEATT[1:0] - Ignored
	*/
	WM8731_AAPC = (0x04 << 1),
	/**	0x5 Digital Audio Path Control -
			   bit 0   = 0  : ADCHPD     - 0 = Enable High Pass Filter
			   bit 2:1 = 00 : DEEMP[1:0] - De-emphasis Control 00 = Disable
			   bit 3   = 0  : DACMU      - 0 = Disable Soft Mute
			   bit 4   = 0  : HPOR       - 0 = Clear Offset
	*/
	WM8731_DAPC = (0x05 << 1),
	/**	0x6 Power Down Control -
			   bit 0 = 1 : LINEINPD - 1 = Enable Power Down
			   bit 1 = 1 : MICPD    - 1 = Enable Power Down
			   bit 2 = 0 : ADCPD    - 0 = Disable Power Down
			   bit 3 = 0 : DACPD    - 0 = Disable Power Down
			   bit 4 = 0 : OUTPD    - 0 = Disable Power Down
			   bit 5 = 0 : OSCPD    - 0 = Disable Power Down
			   bit 6 = 1 : CLKOUTPD - 1 = Enable Power Down
			   bit 7 = 0 : POWEROFF - 0 = Disable POWEROFF
	*/
	WM8731_PWRCTL = (0x06 << 1),
	/**	0x7 Digital Audio Interface Format -
			   bit 1:0 = 10 : FORMAT[1:0] - 10 = I2S
			   bit 3:2 = xx : IWL[1:0]    - 00 = 16bit
										  - 10 = 24 bit
										  - 11 = 32 bit
			   bit 4   = 0  : LRP         - 0 = Right Ch when DACLRC Low
			   bit 5   = 0  : LRSWAP      - 0 = Disable DAC Left Right Clock Swap
			   bit 6   = 1  : MS          - 1 = Enable Master Mode
			   bit 7   = 0  : BCLKINV     - 0 = Don't Invert BCLK
	*/
	WM8731_DAIF = (0x07 << 1),
	/**	0x8 Sampling Control -
			   bit 0   = 1    : USB/NORMAL - 1 = USB Mode (250/272 fs)
			   bit 1   = 0    : BOSR       - 0 = 250 fs (for 48000 Hz sampling)
										   - 1 = 272 fs (for 44100 Hz sampling)
			   bit 5:2 = 0000 : SR[3:0]    - 0000 = 48 kHz sample rate
										   - 1000 = 44.1 kHz sample rate
			   bit 6   = 0    : CLKIDIV2   - 0 = Core Clock is MCLK
			   bit 7   = 0    : CLKODIV2   - 0 = CLOCKOUT is Core
	*/
	WM8731_SAMPLING = (0x08 << 1),
	/**	0x9 Active Control -
			   bit 0 = 1 : ACTIVE - 1 = Active
	*/
	WM8731_ACTIVE = (0x09 << 1),
	/**	0xF Reset -
			 Writing 0x00 to register resets device
	*/
	WM8731_RESET = (0x0F << 1),
};

/** @brief Initial register setup for WM8731 device.
*/
static const reg_t reg_init[] =
{
		{WM8731_PWRCTL, 0x43},
#ifndef AUD_MASTER_MODE_FT90x
#if AUD_BIT_RESOLUTION == 24
		{WM8731_DAIF, 0x4a}, // 24 bit
#elif AUD_BIT_RESOLUTION == 32
		{WM8731_DAIF, 0x4e}, // 32 bit
#elif AUD_BIT_RESOLUTION == 16
		{WM8731_DAIF, 0x42}, // 16 bit
#endif
#endif // AUD_MASTER_MODE_FT90x
		{WM8731_DAPC, 0x00},
		{WM8731_AAPC, 0x13},
#if AUD_SAMPLE_FREQ == 48000
		{WM8731_SAMPLING, 0x01},
#elif AUD_SAMPLE_FREQ == 44100
		{WM8731_SAMPLING, 0x23},
#endif
		{WM8731_ACTIVE, 0x01},
		{0xff, 0xff},
};


#endif // I2S_TYPE
//@}

/** @brief I2S buffer streaming stop flag
 * @details Communicates between I2S interrupt handler and main code
 * to stop streaming as soon as the I2S buffer empties.
 */
volatile int8_t aud_stop_streaming = 0;

static const char* I2C_statusbits[] = {
    "I2C_BUSY (The Bus is currently busy transmitting/receiving)",
    "I2C_ERR (An error occurred (ADDR_ACK, DATA_ACK, ARB_LOST))",
    "ADDR_ACK (Slave address was not acknowledged)",
    "DATA_ACK (Data was not acknowledged)",
    "ARB_LOST (Arbitration lost)",
    "I2C_IDLE (The I2C Controller is idle)",
    "BUS_BUSY (The I2C Bus is busy)",
    "-"
};

/* MACROS **************************************************************************/

/* LOCAL FUNCTIONS / INLINES *******************************************************/

/** @name tfp_putc
 *  @details Machine dependent putc function for tfp_printf (tinyprintf) library.
 *  @param p Parameters (machine dependent)
 *  @param c The character to write
 */
void tfp_putc(void* p, char c)
{
	uart_write((ft900_uart_regs_t*)p, (uint8_t)c);
}

/** @brief Returns the current millisecond country
 *  @returns A count of milliseconds
 */
uint32_t millis(void)
{
	return milliseconds;
}

void timer_ISR(void)
{
	if (timer_is_interrupted(timer_select_a))
	{
		milliseconds++;
	}
}

/* Power management ISR */
void powermanagement_ISR(void)
{
	if (SYS->PMCFG_H & MASK_SYS_PMCFG_HOST_RST_DEV)
	{
		// Clear Host Reset interrupt
		SYS->PMCFG_H = MASK_SYS_PMCFG_HOST_RST_DEV;
		USBD_wakeup();
	}

	if (SYS->PMCFG_H & MASK_SYS_PMCFG_HOST_RESUME_DEV)
	{
		// Clear Host Resume interrupt
		SYS->PMCFG_H = MASK_SYS_PMCFG_HOST_RESUME_DEV;
		USBD_resume();
	}
}

/**
 @brief      USB Set/Get Interface request handler
 @details    Handle standard requests from the host application
 for GetInterface and SetInterface requests.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - USBD_OK if successful or USBD_ERR_*
 if there is an error or the bRequest is not
 supported.
 **/
//@{
int8_t setif_req_cb(USB_device_request *req)
{
	int8_t status = USBD_ERR_NOT_SUPPORTED;

	if (req->wIndex == 0)
	{
		// Interface 0 can only have an Alt Setting of zero.
		if (req->wValue == 0x00)
		{
			status = USBD_OK;
		}
	}
	else if (req->wIndex == 1)
	{
		// Interface 1 can change Alt Settings to zero or 1 in ISOC mode.
		if ((req->wValue == 0x00) || (req->wValue == 0x01))
		{
			usb_alt = LSB(req->wValue);
			status = USBD_OK;

			// Start or stop the dac depending on the alternate interface.
			dac_state_change = 1;
		}
	}

	return status;
}

int8_t getif_req_cb(USB_device_request *req, uint8_t *val)
{
	int8_t status = USBD_ERR_NOT_SUPPORTED;

	// Only interface 1 has Alt Settings in this application.
	if (req->wIndex == 0)
	{
		// Alt setting for interface 0 is always zero.
		*val = 0;
		status = USBD_OK;
	}
	else if (req->wIndex == 1)
	{
		// Alt setting for interface 1 can be changed.
		*val = usb_alt;
		status = USBD_OK;
	}

	return status;
}
//@}

/**
 @brief      USB DFU class request handler (runtime)
 @details    Handle class requests from the host application.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - USBD_OK if successful or USBD_ERR_*
 if there is an error or the bRequest is not
 supported.
 **/
int8_t class_req_dfu_interface_runtime(USB_device_request *req)
{
	int8_t status = USBD_ERR_NOT_SUPPORTED;

	// Handle only DFU_DETATCH, DFU_GETSTATE and DFU_GETSTATUS
	// when in Runtime mode. Table 3.2 DFU_DETACH is mandatory
	// in Runtime mode, DFU_GETSTATE and DFU_GETSTATUS are
	// optional.
	switch (req->bRequest)
	{
	case USB_CLASS_REQUEST_DETACH:
		USBD_DFU_class_req_detach(req->wValue);
		status = USBD_OK;
		break;
	case USB_CLASS_REQUEST_GETSTATUS:
		USBD_DFU_class_req_getstatus(req->wLength);
		status = USBD_OK;
		break;
	case USB_CLASS_REQUEST_GETSTATE:
		USBD_DFU_class_req_getstate(req->wLength);
		status = USBD_OK;
		break;
	}
	return status;
}

/**
 @brief      USB DFU class request handler (DFU mode)
 @details    Handle class requests from the host application.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - USBD_OK if successful or USBD_ERR_*
 if there is an error or the bRequest is not
 supported.
 **/
int8_t class_req_dfu_interface_dfumode(USB_device_request *req)
{
	int8_t status = USBD_ERR_NOT_SUPPORTED;

	// Handle remaining DFU class requests when in DFU Mode.
	// Table 3.2 DFU_DETACH is not supported in DFU Mode.
	switch (req->bRequest)
	{
	case USB_CLASS_REQUEST_DNLOAD:
		/* Block number passed in wValue gives the start address of
		 * to program based on the size of the transfer size.
		 */
		USBD_DFU_class_req_download(req->wValue * DFU_TRANSFER_SIZE,
				req->wLength);
		status = USBD_OK;
		break;

	case USB_CLASS_REQUEST_UPLOAD:
		/* Block number passed in wValue gives the start address of
		 * to program based on the size of the transfer size.
		 */
		USBD_DFU_class_req_upload(req->wValue * DFU_TRANSFER_SIZE,
				req->wLength);
		status = USBD_OK;
		break;

	case USB_CLASS_REQUEST_GETSTATUS:
		USBD_DFU_class_req_getstatus(req->wLength);
		status = USBD_OK;
		break;

	case USB_CLASS_REQUEST_GETSTATE:
		USBD_DFU_class_req_getstate(req->wLength);
		status = USBD_OK;
		break;
	case USB_CLASS_REQUEST_CLRSTATUS:
		USBD_DFU_class_req_clrstatus();
		status = USBD_OK;
		break;
	case USB_CLASS_REQUEST_ABORT:
		USBD_DFU_class_req_abort();
		status = USBD_OK;
		break;

	default:
		// Unknown or unsupported request.
		break;
	}
	return status;
}

int8_t class_req_interface_audio_control(USB_device_request *req)
{
	int8_t status = USBD_ERR_NOT_SUPPORTED;
	uint8_t request = req->bRequest;
	uint8_t requestType = req->bmRequestType;
	//uint8_t controlSelector = MSB(req->wValue);
	//uint8_t entityID = MSB(req->wIndex);

	// Interface requests to the AudioControl interface
	// of the audio function.
	// Interface number is in LSB of wIndex
	if ((requestType & USB_BMREQUESTTYPE_DIR_MASK) ==
			USB_BMREQUESTTYPE_DIR_HOST_TO_DEV)
	{
		switch (request)
		{
		}
	}
	else
	{
		// Interface requests to the AudioControl interface
		// of the audio function.
		// Interface number is in LSB of wIndex
		switch (request)
		{
		}

		if (status == USBD_OK)
		{
			USBD_transfer_ep0(USBD_DIR_OUT, NULL, 0, 0);
		}
	}

	return status;
}


int8_t class_req_interface_audio_streaming(USB_device_request *req)
{
	int8_t status = USBD_ERR_NOT_SUPPORTED;
	uint8_t request = req->bRequest;
	uint8_t requestType = req->bmRequestType;
	//uint8_t controlSelector = MSB(req->wValue);

	// Interface requests to the AudioStreaming interface
	// of the audio function.
	// Interface number is in LSB of wIndex
	if ((requestType & USB_BMREQUESTTYPE_DIR_MASK) ==
			USB_BMREQUESTTYPE_DIR_HOST_TO_DEV)
	{
		switch (request)
		{
		}
	}
	else
	{
		// Interface requests to the AudioStreaming interface
		// of the audio function.
		// Interface number is in LSB of wIndex

		switch (request)
		{
		}

		USBD_transfer_ep0(USBD_DIR_OUT, NULL, 0, 0);
		status = USBD_OK;
	}

	return status;
}

int8_t class_req_endpoint_audio(USB_device_request *req)
{
	int8_t status = USBD_ERR_NOT_SUPPORTED;
	//uint8_t command[USB_CONTROL_EP_MAX_PACKET_SIZE];
	uint8_t request = req->bRequest;

	// Endpoint requests to the AudioStreaming endpoint
	// of the audio function.
	// Endpoint number is in LSB of wIndex
	// Entity ID is MSB of wIndex
	switch (request)
	{
	}

	return status;
}

/**
 @brief      USB class request handler
 @details    Handle class requests from the host application.
 The bRequest value is parsed and the appropriate
 action or function is performed. Additional values
 from the USB_device_request structure are decoded
 and the DFU state machine and status updated as
 required.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - USBD_OK if successful or USBD_ERR_*
 if there is an error or the bRequest is not
 supported.
 **/
int8_t class_req_cb(USB_device_request *req)
{
	int8_t status = USBD_ERR_NOT_SUPPORTED;
	uint8_t requestType = req->bmRequestType;
	uint8_t interface = LSB(req->wIndex) & 0x0F;

	// For DFU requests the SETUP packet must be addressed to the
	// the DFU interface on the device.

	// For DFU requests ensure the recipient is an interface...
	if ((requestType & USB_BMREQUESTTYPE_RECIPIENT_MASK) ==
			USB_BMREQUESTTYPE_RECIPIENT_INTERFACE)
	{
		// ...and that the interface is the correct Runtime interface
		if (USBD_DFU_is_runtime())
		{
			if ((interface == DFU_USB_INTERFACE_RUNTIME))
			{
				status = class_req_dfu_interface_runtime(req);
			}
			else if (interface == 0)
			{
				status = class_req_interface_audio_control(req);
			}
			else if (interface == 1)
			{
				status = class_req_interface_audio_streaming(req);
			}
		}
		// ...or the correct DFU Mode interface
		else
		{
			if (interface == DFU_USB_INTERFACE_DFUMODE)
			{
				status = class_req_dfu_interface_dfumode(req);
			}
		}
	}
	else if ((requestType & USB_BMREQUESTTYPE_RECIPIENT_MASK) ==
			USB_BMREQUESTTYPE_RECIPIENT_ENDPOINT)
	{
		status = class_req_endpoint_audio(req);
	}

	return status;
}

/**
 @brief      USB standard request handler for GET_DESCRIPTOR
 @details    Handle standard GET_DESCRIPTOR requests from the host
 application.
 The hValue is parsed and the appropriate device,
 configuration or string descriptor is selected.
 For device and configuration descriptors the DFU
 state machine determines which descriptor to use.
 For string descriptors the lValue selects which
 string from the string_descriptors table to use.
 The string table is automatically traversed to find
 the correct string and the length is taken from the
 string descriptor.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - USBD_OK if successful or USBD_ERR_*
 if there is an error or the bmRequestType is not
 supported.
 **/
int8_t standard_req_get_descriptor(USB_device_request *req, uint8_t **buffer, uint16_t *len)
{
	uint8_t *src = NULL;
	uint16_t length = req->wLength;
	uint8_t hValue = MSB(req->wValue);
	uint8_t lValue = LSB(req->wValue);
	uint8_t i;
	uint8_t slen;

	switch (hValue)
	{
	case USB_DESCRIPTOR_TYPE_DEVICE:
		if (USBD_DFU_is_runtime())
		{
			src = (uint8_t *) &device_descriptor_uvc;
		}
		else
		{
			src = (uint8_t *) &device_descriptor_dfumode;
		}
		if (length > sizeof(USB_device_descriptor)) // too many bytes requested
			length = sizeof(USB_device_descriptor); // Entire structure.
		break;

	case USB_DESCRIPTOR_TYPE_CONFIGURATION:
		if (USBD_DFU_is_runtime())
		{
			if (USBD_get_bus_speed() == USBD_SPEED_HIGH)
			{
				memcpy((void *)&config_descriptor_buffer.hs, (void *)&config_descriptor_uad_hs, sizeof(config_descriptor_uad_hs));
				if (length > sizeof(config_descriptor_uad_hs)) // too many bytes requested
					length = sizeof(config_descriptor_uad_hs); // Entire structure.
			}
			else
			{
				memcpy((void *)&config_descriptor_buffer.fs, (void *)&config_descriptor_uad_fs, sizeof(config_descriptor_uad_fs));
				if (length > sizeof(config_descriptor_uad_fs)) // too many bytes requested
					length = sizeof(config_descriptor_uad_fs); // Entire structure.
			}
			src = (uint8_t *)&config_descriptor_buffer.hs;
			config_descriptor_buffer.hs.configuration.bDescriptorType = USB_DESCRIPTOR_TYPE_CONFIGURATION;
		}
		else
		{
			src = (uint8_t *) &config_descriptor_dfumode;
			if (length > sizeof(config_descriptor_dfumode)) // too many bytes requested
				length = sizeof(config_descriptor_dfumode); // Entire structure.
		}
		break;

	case USB_DESCRIPTOR_TYPE_OTHER_SPEED_CONFIGURATION:
		if (USBD_DFU_is_runtime())
		{
			if (USBD_get_bus_speed() == USBD_SPEED_HIGH)
			{
				memcpy((void *)&config_descriptor_buffer.fs, (void *)&config_descriptor_uad_fs, sizeof(config_descriptor_uad_fs));
				if (length > sizeof(config_descriptor_uad_fs)) // too many bytes requested
					length = sizeof(config_descriptor_uad_fs); // Entire structure.
			}
			else
			{
				memcpy((void *)&config_descriptor_buffer.hs, (void *)&config_descriptor_uad_hs, sizeof(config_descriptor_uad_hs));
				if (length > sizeof(config_descriptor_uad_hs)) // too many bytes requested
					length = sizeof(config_descriptor_uad_hs); // Entire structure.
			}
			src = (uint8_t *)&config_descriptor_buffer.hs;
			config_descriptor_buffer.hs.configuration.bDescriptorType = USB_DESCRIPTOR_TYPE_OTHER_SPEED_CONFIGURATION;
		}
		break;

	case USB_DESCRIPTOR_TYPE_DEVICE_QUALIFIER:
		src = (uint8_t *) &device_qualifier_descriptor_uvc;
		if (length > sizeof(USB_device_qualifier_descriptor)) // too many bytes requested
			length = sizeof(USB_device_qualifier_descriptor); // Entire structure.
		break;

	case USB_DESCRIPTOR_TYPE_STRING:
		// Special case: WCID descriptor
#ifdef USB_INTERFACE_USE_DFU
		if (lValue == USB_MICROSOFT_WCID_STRING_DESCRIPTOR)
		{
			src = (uint8_t *)wcid_string_descriptor;
			length = sizeof(wcid_string_descriptor);
			break;
		}
#endif // USB_INTERFACE_USE_DFU

		// Find the nth string in the string descriptor table
		i = 0;
		while ((slen = ((uint8_t *)string_descriptor)[i]) > 0)
		{
			// Point to start of string descriptor in __code segment
			// Typecast prevents warning for losing const when USBD_transfer
			// is called
			src = (uint8_t *) &((uint8_t *)string_descriptor)[i];
			if (lValue == 0)
			{
				break;
			}
			i += slen;
			lValue--;
		}
		if (lValue > 0)
		{
			// String not found
			return USBD_ERR_NOT_SUPPORTED;
		}
		// Update the length returned only if it is less than the requested
		// size
		if (length > slen)
		{
			length = slen;
		}

		break;

	default:
		return USBD_ERR_NOT_SUPPORTED;
	}

	*buffer = src;
	*len = length;

	return USBD_OK;
}

/**
 @brief      USB vendor request handler
 @details    Handle vendor requests from the host application.
 The bRequest value is parsed and the appropriate
 action or function is performed. Additional values
 from the USB_device_request structure are decoded
 and provided to other handlers.
 There are no vendor requests requiring handling in
 this firmware.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - USBD_OK if successful or USBD_ERR_*
 if there is an error or the bRequest is not
 supported.
 **/
int8_t vendor_req_cb(USB_device_request *req)
{
	int8_t status = USBD_ERR_NOT_SUPPORTED;
#ifdef USB_INTERFACE_USE_DFU
	uint16_t length = req->wLength;
#endif // USB_INTERFACE_USE_DFU

	// For Microsoft WCID only.
	// Request for Compatible ID Feature Descriptors.
	// Check the request code and wIndex.
#ifdef USB_INTERFACE_USE_DFU
	if (req->bRequest == WCID_VENDOR_REQUEST_CODE)
	{
		if (req->wIndex == USB_MICROSOFT_WCID_FEATURE_WINDEX_COMPAT_ID)
		{
			if (req->bmRequestType & USB_BMREQUESTTYPE_DIR_DEV_TO_HOST)
			{
				if (length > sizeof(wcid_feature_runtime)) // too many bytes requested
					length = sizeof(wcid_feature_runtime); // Entire structure.
				// Return a compatible ID feature descriptor.
				if (USBD_DFU_is_runtime())
				{
					USBD_transfer_ep0(USBD_DIR_IN, (uint8_t *) &wcid_feature_runtime,
							length, length);
				}
				else
				{
					USBD_transfer_ep0(USBD_DIR_IN, (uint8_t *) &wcid_feature_dfumode,
							length, length);
				}
				// ACK packet
				USBD_transfer_ep0(USBD_DIR_OUT, NULL, 0, 0);
				status = USBD_OK;
			}
		}
	}
#endif // USB_INTERFACE_USE_DFU

	return status;
}

/**
 @brief      USB suspend callback
 @details    Called when the USB bus enters the suspend state.
 @param[in]	status - unused.
 @return		N/A
 **/
void suspend_cb(uint8_t status)
{
	(void) status;

	SYS->PMCFG_L |= MASK_SYS_PMCFG_HOST_RESUME_DEV;

	tfp_printf("Suspend\r\n");
	return;
}

/**
 @brief      USB resume callback
 @details    Called when the USB bus enters the resume state
 prior to restating after a suspend.
 @param[in]  status - unused.
 @return     N/S
 **/
void resume_cb(uint8_t status)
{
	(void) status;

	SYS->PMCFG_L &= (~MASK_SYS_PMCFG_HOST_RESUME_DEV);

	tfp_printf("Resume\r\n");
	return;
}

/**
 @brief      USB reset callback
 @details    Called when the USB bus enters the reset state.
 @param[in]	status - unused.
 @return		N/A
 **/
void reset_cb(uint8_t status)
{
	(void) status;

	USBD_set_state(USBD_STATE_DEFAULT);

	// If we are at DFUSTATE_MANIFEST_WAIT_RESET stage and do
	// not support detach then the DFU will reset the chip and
	// run the new firmware.
	// From the DFUSTATE_APPIDLE state advance to DFUSTATE_DFUIDLE.
	// Other states will not advance the state machine or may
	// move this to DFUSTATE_DFUERROR if it is inappropriate to
	// find a reset at that stage.
	USBD_DFU_reset();
	tfp_printf("Reset\r\n");

	return;
}

int8_t dac_init(void)
{
	int i = 0;
	int8_t retval = 0;
    uint8_t i2c_addr;

#if I2S_TYPE == WM8731
    /* Set up the WM8731 ... */
    tfp_printf("WM8731 Device\r\n");
    i2c_addr = WM8731_WR_ADDR;
#endif

	while (reg_init[i].addr != 0xff)
	{
		retval = i2cm_write(i2c_addr, reg_init[i].addr, &(reg_init[i].reg), 1);
		if (retval)
		{
			uint8_t status = i2cm_get_status();
			int i;

			tfp_printf("i2cm_write() returned %d\r\n", retval);

			for(i=0; i<8; ++i)
			{
				if (status & (1<<i))
				{
					tfp_printf("    %s\r\n",I2C_statusbits[i]);
				}
			}

			break;
		}
		i++;
	}
	delayms(100);

	return retval;
}

/**
 @brief      DAC reset
 @details    Will send a reset to the DAC.
 	 	 	 This function abstracts the underlying DAC hardware.
 @return	 Zero if successful.
 **/
int8_t dac_reset(void)
{
	int8_t retval = 0;

#if I2S_TYPE == WM8731
	reg_t reg_reset = {WM8731_RESET, 0x00};

    retval = i2cm_write(WM8731_WR_ADDR, reg_reset.addr, &(reg_reset.reg), 1);

#endif // I2S_TYPE

    return retval;
}

/**
 @brief      DAC soft mute
 @details    Will send a mute signal to the DAC.
 	 	 	 This function abstracts the underlying DAC hardware.
 @return	 Zero if successful.
 **/
int8_t dac_soft_mute()
{
	int8_t retval = 0;

#if I2S_TYPE == WM8731
	reg_t reg_mute = {WM8731_DAPC, 0x08};

    retval = i2cm_write(WM8731_WR_ADDR, reg_mute.addr, &(reg_mute.reg), 1);

#endif // I2S_TYPE

    return retval;
}

/**
 @brief      DAC soft unmute
 @details    Will send an unmute signal to the DAC.
 	 	 	 This function abstracts the underlying DAC hardware.
 @return	 Zero if successful.
 **/
int8_t dac_soft_unmute()
{
	int8_t retval = 0;

#if I2S_TYPE == WM8731
   	reg_t reg_unmute = {WM8731_DAPC, 0x00};

   	retval = i2cm_write(WM8731_WR_ADDR, reg_unmute.addr, &(reg_unmute.reg), 1);

#endif // I2S_TYPE

    return retval;
}

/* ISR which will stream data to the I2S slave device when it is needed */
void i2s_ISR(void)
{
	uint16_t flag = i2s_get_status();

	/* If the FIFO is empty... */
    if ((aud_stop_streaming) && (flag & MASK_I2S_PEND_FIFO_TX_EMPTY))
    {
		i2s_stop_tx();
		i2s_disable_int(MASK_I2S_IE_FIFO_TX_EMPTY | MASK_I2S_PEND_FIFO_TX_HALF_FULL);
		aud_stop_streaming = 0;
    }
    else if (flag & MASK_I2S_PEND_FIFO_TX_EMPTY)
	{
		i2s_write(dac_zero, AUD_SAMPLE_FILL);
		i2s_clear_int_flag(MASK_I2S_PEND_FIFO_TX_EMPTY);
	}
    else if (flag & MASK_I2S_PEND_FIFO_TX_HALF_FULL)
	{
#ifdef AUD_USE_BUFFER
    	// Send a minimum size of data at a time.
		if (dac_buffer_size > 512)
		{
			// Normal condition. Send all data in buffer to I2S.
			if (dac_rd + dac_buffer_size < AUDIO_DAC_BUFFER)
			{
				i2s_write(&dac_buffer[dac_rd], dac_buffer_size);
				dac_rd += dac_buffer_size;
			}
			else
			{
				i2s_write(&dac_buffer[dac_rd], AUDIO_DAC_BUFFER - dac_rd);
				dac_buffer_size -= (AUDIO_DAC_BUFFER - dac_rd);
				dac_rd = 0;
				if (dac_buffer_size > 0)
				{
					i2s_write(&dac_buffer[dac_rd], dac_buffer_size);
				}
				dac_rd += dac_buffer_size;
			}
			dac_buffer_size = 0;
		}
		else
#endif // AUD_USE_BUFFER
		{
			i2s_write(dac_zero, AUD_SAMPLE_FILL);
		}
		i2s_clear_int_flag(MASK_I2S_PEND_FIFO_TX_HALF_FULL);
	}

}

uint8_t usbd_testing(void)
{
	USBD_ctx usb_ctx;

	// Packet length.
	uint16_t packet_len;

	memset(&usb_ctx, 0, sizeof(usb_ctx));

	usb_ctx.standard_req_cb = NULL;
	usb_ctx.get_descriptor_cb = standard_req_get_descriptor;
	usb_ctx.set_interface_cb = setif_req_cb;
	usb_ctx.get_interface_cb = getif_req_cb;
	usb_ctx.class_req_cb = class_req_cb;
	usb_ctx.vendor_req_cb = vendor_req_cb;
	usb_ctx.suspend_cb = suspend_cb;
	usb_ctx.resume_cb = resume_cb;
	usb_ctx.reset_cb = reset_cb;
	usb_ctx.lpm_cb = NULL;
	usb_ctx.speed = USBD_SPEED_HIGH;

	// Initialise the USB device with a control endpoint size
	// of 8 to 64 bytes. This must match the size for bMaxPacketSize0
	// defined in the device descriptor.
	usb_ctx.ep0_size = USB_CONTROL_EP_SIZE;
	//usb_ctx.ep0_cb = ep_cb;

	USBD_initialise(&usb_ctx);

	// Wait for connect to host.
	while (USBD_connect() != USBD_OK);

	if (USBD_get_bus_speed() == USBD_SPEED_HIGH)
	{
		USBD_create_endpoint(UAD_EP_AUDIO_STREAM, USBD_EP_ISOC, USBD_DIR_OUT,
				UAD_EP_AUDIO_STREAM_USBD_SIZE, USBD_DB_ON, NULL /*ep_cb*/);

		packet_len = AUD_PACKET_LEN_HS;
	}
	else
	{
		USBD_create_endpoint(UAD_EP_AUDIO_STREAM, USBD_EP_ISOC, USBD_DIR_OUT,
				UAD_EP_AUDIO_STREAM_USBD_SIZE, USBD_DB_ON, NULL /*ep_cb*/);

		packet_len = AUD_PACKET_LEN_FS;
	}

	tfp_printf("Starting packet length %d\r\n", packet_len);

	dac_state_change = 0;

	for (;;)
	{
		if (USBD_connect() == USBD_OK)
		{
			if (USBD_DFU_is_runtime())
			{
				// Start the UVC emulation code.
				while (USBD_process() == USBD_OK)
				{
					if (dac_state_change)
					{
						dac_state_change = 0;
						// Start or stop the AUDIO depending on the alternate interface.
						if (usb_alt == 1)
						{
							tfp_printf("Audio starting\r\n");
						    //i2s_clear_int_flag(0xFFFF);
						    memset(dac_buffer, 0, sizeof(dac_buffer));
						    memset(dac_zero, 0, sizeof(dac_zero));

						    {
						    	uint16_t len;

						    	CRITICAL_SECTION_BEGIN
						    	// Seed the I2S buffer
						    	for (len = 0; len < (AUDIO_DAC_BUFFER / 2) + 16; len += AUD_SAMPLE_FILL)
						    	{
						    		i2s_write(dac_zero, AUD_SAMPLE_FILL);
						    	}
						    	aud_stop_streaming = 0;
#ifdef AUD_USE_BUFFER
						    	dac_buffer_size = 0;
						    	dac_wr = 0;
						    	dac_rd = 0;
#endif // AUD_USE_BUFFER
						    	CRITICAL_SECTION_END
						    }
						    i2s_enable_int(MASK_I2S_IE_FIFO_TX_EMPTY | MASK_I2S_IE_FIFO_TX_HALF_FULL);

						    i2s_start_tx();

							dac_soft_unmute();
						}
						else
						{
							dac_soft_mute();
							CRITICAL_SECTION_BEGIN
						    aud_stop_streaming = 1;
							CRITICAL_SECTION_END
							tfp_printf("Audio stopped\r\n");
						}
					}

					if (usb_alt == 1)
					{
						do {
							if (USBD_ep_buffer_full(UAD_EP_AUDIO_STREAM))
							{
								uint16_t len;
#ifdef AUD_USE_BUFFER
								uint8_t *p_wr;
								uint8_t *p_rd = recv_packet;
#endif // AUD_USE_BUFFER

								len = USBD_transfer_ex(UAD_EP_AUDIO_STREAM,
										recv_packet,
										packet_len,
										0, 0);

								if (len > 0)
								{
									CRITICAL_SECTION_BEGIN
#ifdef AUD_USE_BUFFER
									// Pointer to the write location in the buffer.
									p_wr = &dac_buffer[dac_wr];

									if (dac_wr + len < AUDIO_DAC_BUFFER)
									{
										// fast memcpy
										__asm__("memcpy.l %0,%1,%2" : :"r"(p_wr), "r"(p_rd), "r"(len));

										dac_wr += len;
									}
									else
									{
										uint16_t len1 = AUDIO_DAC_BUFFER - dac_wr;
										uint16_t len2 = len - len1;

										// fast memcpy
										__asm__("memcpy.l %0,%1,%2" : :"r"(p_wr), "r"(p_rd), "r"(len1));

										p_wr = &dac_buffer[0];
										p_rd += len1;

										if (len2)
										{
											__asm__("memcpy.l %0,%1,%2" : :"r"(p_wr), "r"(p_rd), "r"(len2));
										}

										dac_wr = len2;
									}

									dac_buffer_size += len;
#else // AUD_USE_BUFFER
									i2s_write(recv_packet, len);
#endif // AUD_USE_BUFFER

									CRITICAL_SECTION_END
								}
							}

						} while (0);
					}
				}
			}
		}
		else
		{
			// In DFU mode. Process USB requests.
			while (USBD_process() == USBD_OK);
		}
		tfp_printf("Restarting\r\n");
	}

	return 0;
}

/* FUNCTIONS ***********************************************************************/

int main(void)
{
	int8_t retval = 0;

#ifdef USB_INTERFACE_USE_STARTUPDFU
	STARTUP_DFU();
#endif // USB_INTERFACE_USE_STARTUPDFU

	sys_reset_all();
	sys_disable(sys_device_i2s);
	sys_disable(sys_device_i2c_master);
	sys_disable(sys_device_usb_device);

	/* Enable the UART Device... */
	sys_enable(sys_device_uart0);
	/* Make GPIO48 function as UART0_TXD and GPIO49 function as UART0_RXD... */
	gpio_function(48, pad_uart0_txd); /* UART0 TXD */
	gpio_function(49, pad_uart0_rxd); /* UART0 RXD */

	/* Enable tfp_printf() functionality... */
	init_printf(UART0, tfp_putc);

#ifdef BRIDGE_DEBUG
	// Open the UART using the coding required.
	uart_open(UART0, 1, UART_DIVIDER_19200_BAUD, uart_data_bits_8, uart_parity_none, uart_stop_bits_1);

	/* Print out a welcome message... */
	BRIDGE_DEBUG_PRINTF("\x1B[2J" /* ANSI/VT100 - Clear the Screen */
			"\x1B[H"  /* ANSI/VT100 - Move Cursor to Home */
	);

	BRIDGE_DEBUG_PRINTF("(C) Copyright 2016, Bridgetek Pte Ltd. \r\n");
	BRIDGE_DEBUG_PRINTF("--------------------------------------------------------------------- \r\n");
	BRIDGE_DEBUG_PRINTF("Welcome to the USB Audio Device  ... \r\n");
	BRIDGE_DEBUG_PRINTF("\r\n");
	BRIDGE_DEBUG_PRINTF("Emulate a USB Audio Device connected to the USB.\r\n");
	BRIDGE_DEBUG_PRINTF("--------------------------------------------------------------------- \r\n");
#endif // BRIDGE_DEBUG

	/* Timer A = 1ms */
	timer_prescaler(1000);
	timer_init(timer_select_a, 100, timer_direction_down, timer_prescaler_select_on, timer_mode_continuous);
	timer_enable_interrupt(timer_select_a);
	timer_start(timer_select_a);

	interrupt_attach(interrupt_timers, (int8_t)interrupt_timers, timer_ISR);
	/* Enable power management interrupts. Primarily to detect resume signalling
	 * from the USB host. */
	interrupt_attach(interrupt_0, (int8_t)interrupt_0, powermanagement_ISR);

    /* MM900EV2: Bring the output amplifiers out of Power Down ... */
    gpio_dir(65, pad_dir_output);
    gpio_write(65, 1);

    /* Set up the I2S peripheral, make sure GPIO 62 and 63 are set up for I2S Slave ... */
    sys_enable(sys_device_i2s);
    gpio_function(60, pad_i2s_sdao); gpio_pull(60, pad_pull_none); /* I2S SDAO */
    gpio_function(61, pad_i2s_sdai); gpio_pull(61, pad_pull_none); /* I2S SDAI */
#ifndef AUD_MASTER_MODE_FT90x
    gpio_function(62, pad_i2ss_bclk); gpio_pull(62, pad_pull_none); /* I2S Slave BCLK */
    gpio_function(63, pad_i2ss_lrclk); gpio_pull(63, pad_pull_none); /* I2S Slave LRCLK */
#else // AUD_MASTER_MODE_FT90x
    gpio_function(62, pad_i2s_bclk); gpio_pull(62, pad_pull_none); /* I2S Master BCLK */
    gpio_function(63, pad_i2s_lrclk); gpio_pull(63, pad_pull_none); /* I2S Master LRCLK */
#endif // AUD_MASTER_MODE_FT90x
    gpio_function(64, pad_i2s_mclk); gpio_pull(64, pad_pull_none); /* I2S MCLK */
    gpio_function(65, pad_i2s_clk22); gpio_pull(65, pad_pull_none); /* I2S CLK22 */
    gpio_function(66, pad_i2s_clk24); gpio_pull(66, pad_pull_none); /* I2S CLK24 */

    /* The MM900EV2A board can only have the FT90x configured as an I2S Slave
     * and the DAC device as the I2S Master.
     */
    i2s_init(
#ifndef AUD_MASTER_MODE_FT90x
    		i2s_mode_slave,
#else //  AUD_MASTER_MODE_FT90x
    		i2s_mode_master,
#endif // AUD_MASTER_MODE_FT90x

#if AUD_BIT_RESOLUTION == 24
    		i2s_length_24,
#elif AUD_BIT_RESOLUTION == 32
			i2s_length_32,
#elif AUD_BIT_RESOLUTION == 16
			i2s_length_16,
#endif
             i2s_format_i2s,
             i2s_padding_8,
             i2s_master_input_clk_22mhz,
#if AUD_BIT_RESOLUTION == 16
             i2s_bclk_div_16,
			 i2s_mclk_div_2,
			 i2s_bclk_per_channel_16
#else
             i2s_bclk_div_8,
			 i2s_mclk_div_2,
			 i2s_bclk_per_channel_32
#endif
    );

    /* Set up the I2C peripheral ... */
    sys_enable(sys_device_i2c_master);
#if I2C_CHANNEL == 0
    /* Set GPIO44 to I2CM_SCL and GPIO45 to I2CM_SDA... */
    gpio_function(44, pad_i2c0_scl);
    gpio_pull(44, pad_pull_none);
    gpio_function(45, pad_i2c0_sda);
    gpio_pull(45, pad_pull_none);
#elif I2C_CHANNEL == 1
    /* Set GPIO46 to I2CM_SCL and GPIO47 to I2CM_SDA... */
    gpio_function(44, pad_i2c0_scl);
    gpio_pull(46, pad_pull_none);
    gpio_function(45, pad_i2c0_sda);
    gpio_pull(47, pad_pull_none);
    /* Set the I2C Master pins to channel 1 */
    sys_i2c_swop(1);
#endif // I2C_CHANNEL

    i2cm_init(I2CM_NORMAL_SPEED, 10000);

	dac_reset();
	dac_init();

    /* Set up the ISR for the I2S device... */
	i2s_stop_tx();
	i2s_clear_int_flag(0xFFFF);
    interrupt_attach(interrupt_i2s, (uint8_t)interrupt_i2s, i2s_ISR);

    i2cm_init(I2CM_NORMAL_SPEED, 100000);

	if (retval == 0) // i2s device setup failed
	{
		interrupt_enable_globally();

		usbd_testing();

		interrupt_disable_globally();
	}
	else
	{
		tfp_printf("DAC not found\n");
	}

	// Wait forever...
	for (;;);

	return 0;
}
